package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.parser.Parser;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.codegen.ConstantPool;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.problem.AbortMethod;
import org.eclipse.jdt.internal.compiler.flow.ExceptionHandlingFlowContext;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.InitializationFlowContext;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;

public class Clinit extends AbstractMethodDeclaration
{
    private static int ENUM_CONSTANTS_THRESHOLD;
    private FieldBinding assertionSyntheticFieldBinding;
    private FieldBinding classLiteralSyntheticField;
    
    static {
        Clinit.ENUM_CONSTANTS_THRESHOLD = 2000;
    }
    
    public Clinit(CompilationResult compilationResult) {
        super(compilationResult);
        this.assertionSyntheticFieldBinding = null;
        this.classLiteralSyntheticField = null;
        this.modifiers = 0;
        this.selector = TypeConstants.CLINIT;
    }
    
    public void analyseCode(ClassScope classScope, InitializationFlowContext staticInitializerFlowContext, FlowInfo flowInfo) {
        if (this.ignoreFurtherInvestigation) {
            return;
        }
        try {
            ExceptionHandlingFlowContext clinitContext = new ExceptionHandlingFlowContext(staticInitializerFlowContext.parent, this, Binding.NO_EXCEPTIONS, staticInitializerFlowContext, this.scope, FlowInfo.DEAD_END);
            if ((flowInfo.tagBits & 0x1) == 0x0) {
                this.bits |= 0x40;
            }
            flowInfo = flowInfo.mergedWith(staticInitializerFlowContext.initsOnReturn);
            FieldBinding[] fields = this.scope.enclosingSourceType().fields();
            for (int i = 0, count = fields.length; i < count; ++i) {
                FieldBinding field = fields[i];
                if (field.isStatic() && !flowInfo.isDefinitelyAssigned(field)) {
                    if (field.isFinal()) {
                        this.scope.problemReporter().uninitializedBlankFinalField(field, this.scope.referenceType().declarationOf(field.original()));
                    }
                    else if (field.isNonNull()) {
                        this.scope.problemReporter().uninitializedNonNullField(field, this.scope.referenceType().declarationOf(field.original()));
                    }
                }
            }
            staticInitializerFlowContext.checkInitializerExceptions(this.scope, clinitContext, flowInfo);
        }
        catch (AbortMethod abortMethod) {
            this.ignoreFurtherInvestigation = true;
        }
        return;
    }
    
    @Override
    public void generateCode(ClassScope classScope, ClassFile classFile) {
        int clinitOffset = 0;
        if (this.ignoreFurtherInvestigation) {
            return;
        }
        CompilationResult unitResult = null;
        int problemCount = 0;
        if (classScope != null) {
            TypeDeclaration referenceContext = classScope.referenceContext;
            if (referenceContext != null) {
                unitResult = referenceContext.compilationResult();
                problemCount = unitResult.problemCount;
            }
        }
        boolean restart = false;
        while (true) {
            try {
                clinitOffset = classFile.contentsOffset;
                this.generateCode(classScope, classFile, clinitOffset);
                restart = false;
            }
            catch (AbortMethod e) {
                if (e.compilationResult == CodeStream.RESTART_IN_WIDE_MODE) {
                    classFile.contentsOffset = clinitOffset;
                    --classFile.methodCount;
                    classFile.codeStream.resetInWideMode();
                    if (unitResult != null) {
                        unitResult.problemCount = problemCount;
                    }
                    restart = true;
                }
                else if (e.compilationResult == CodeStream.RESTART_CODE_GEN_FOR_UNUSED_LOCALS_MODE) {
                    classFile.contentsOffset = clinitOffset;
                    --classFile.methodCount;
                    classFile.codeStream.resetForCodeGenUnusedLocals();
                    if (unitResult != null) {
                        unitResult.problemCount = problemCount;
                    }
                    restart = true;
                }
                else {
                    classFile.contentsOffset = clinitOffset;
                    --classFile.methodCount;
                    restart = false;
                }
            }
            if (!restart) {
                break;
            }
            continue;
        }
        return;
    }
    
    private void generateCode(ClassScope classScope, ClassFile classFile, int clinitOffset) {
        ConstantPool constantPool = classFile.constantPool;
        int constantPoolOffset = constantPool.currentOffset;
        int constantPoolIndex = constantPool.currentIndex;
        classFile.generateMethodInfoHeaderForClinit();
        int codeAttributeOffset = classFile.contentsOffset;
        classFile.generateCodeAttributeHeader();
        CodeStream codeStream = classFile.codeStream;
        this.resolve(classScope);
        codeStream.reset(this, classFile);
        TypeDeclaration declaringType = classScope.referenceContext;
        MethodScope staticInitializerScope = declaringType.staticInitializerScope;
        staticInitializerScope.computeLocalVariablePositions(0, codeStream);
        if (this.assertionSyntheticFieldBinding != null) {
            codeStream.generateClassLiteralAccessForType(classScope.outerMostClassScope().enclosingSourceType(), this.classLiteralSyntheticField);
            codeStream.invokeJavaLangClassDesiredAssertionStatus();
            BranchLabel falseLabel = new BranchLabel(codeStream);
            codeStream.ifne(falseLabel);
            codeStream.iconst_1();
            BranchLabel jumpLabel = new BranchLabel(codeStream);
            codeStream.decrStackSize(1);
            codeStream.goto_(jumpLabel);
            falseLabel.place();
            codeStream.iconst_0();
            jumpLabel.place();
            codeStream.fieldAccess((byte)(-77), this.assertionSyntheticFieldBinding, null);
        }
        FieldDeclaration[] fieldDeclarations = declaringType.fields;
        int sourcePosition = -1;
        int remainingFieldCount = 0;
        if (TypeDeclaration.kind(declaringType.modifiers) == 3) {
            int enumCount = declaringType.enumConstantsCounter;
            if (enumCount > Clinit.ENUM_CONSTANTS_THRESHOLD) {
                int begin = -1;
                int count = 0;
                if (fieldDeclarations != null) {
                    int max3 = fieldDeclarations.length;
                    for (int k = 0; k < max3; ++k) {
                        FieldDeclaration fieldDecl3 = fieldDeclarations[k];
                        if (fieldDecl3.isStatic()) {
                            if (fieldDecl3.getKind() == 3) {
                                if (begin == -1) {
                                    begin = k;
                                }
                                if (++count > Clinit.ENUM_CONSTANTS_THRESHOLD) {
                                    SyntheticMethodBinding syntheticMethod = declaringType.binding.addSyntheticMethodForEnumInitialization(begin, k);
                                    codeStream.invoke((byte)(-72), syntheticMethod, null);
                                    begin = k;
                                    count = 1;
                                }
                            }
                            else {
                                ++remainingFieldCount;
                            }
                        }
                    }
                    if (count != 0) {
                        SyntheticMethodBinding syntheticMethod2 = declaringType.binding.addSyntheticMethodForEnumInitialization(begin, max3);
                        codeStream.invoke((byte)(-72), syntheticMethod2, null);
                    }
                }
            }
            else if (fieldDeclarations != null) {
                for (int j = 0, max2 = fieldDeclarations.length; j < max2; ++j) {
                    FieldDeclaration fieldDecl2 = fieldDeclarations[j];
                    if (fieldDecl2.isStatic()) {
                        if (fieldDecl2.getKind() == 3) {
                            fieldDecl2.generateCode(staticInitializerScope, codeStream);
                        }
                        else {
                            ++remainingFieldCount;
                        }
                    }
                }
            }
            codeStream.generateInlinedValue(enumCount);
            codeStream.anewarray(declaringType.binding);
            if (enumCount > 0 && fieldDeclarations != null) {
                for (int j = 0, max2 = fieldDeclarations.length; j < max2; ++j) {
                    FieldDeclaration fieldDecl2 = fieldDeclarations[j];
                    if (fieldDecl2.getKind() == 3) {
                        codeStream.dup();
                        codeStream.generateInlinedValue(fieldDecl2.binding.id);
                        codeStream.fieldAccess((byte)(-78), fieldDecl2.binding, null);
                        codeStream.aastore();
                    }
                }
            }
            codeStream.fieldAccess((byte)(-77), declaringType.enumValuesSyntheticfield, null);
            if (remainingFieldCount != 0) {
                int j = 0;
                int max2 = fieldDeclarations.length;
                while (j < max2) {
                    if (remainingFieldCount >= 0) {
                        FieldDeclaration fieldDecl2 = fieldDeclarations[j];
                        switch (fieldDecl2.getKind()) {
                            case 3: {
                                break;
                            }
                            case 2: {
                                if (!fieldDecl2.isStatic()) {
                                    break;
                                }
                                --remainingFieldCount;
                                sourcePosition = ((Initializer)fieldDecl2).block.sourceEnd;
                                fieldDecl2.generateCode(staticInitializerScope, codeStream);
                                break;
                            }
                            case 1: {
                                if (!fieldDecl2.binding.isStatic()) {
                                    break;
                                }
                                --remainingFieldCount;
                                sourcePosition = fieldDecl2.declarationEnd;
                                fieldDecl2.generateCode(staticInitializerScope, codeStream);
                                break;
                            }
                        }
                        ++j;
                        continue;
                    }
                    break;
                }
            }
        }
        else if (fieldDeclarations != null) {
            for (int i = 0, max = fieldDeclarations.length; i < max; ++i) {
                FieldDeclaration fieldDecl = fieldDeclarations[i];
                switch (fieldDecl.getKind()) {
                    case 2: {
                        if (!fieldDecl.isStatic()) {
                            break;
                        }
                        sourcePosition = ((Initializer)fieldDecl).block.sourceEnd;
                        fieldDecl.generateCode(staticInitializerScope, codeStream);
                        break;
                    }
                    case 1: {
                        if (!fieldDecl.binding.isStatic()) {
                            break;
                        }
                        sourcePosition = fieldDecl.declarationEnd;
                        fieldDecl.generateCode(staticInitializerScope, codeStream);
                        break;
                    }
                }
            }
        }
        if (codeStream.position == 0) {
            classFile.contentsOffset = clinitOffset;
            --classFile.methodCount;
            constantPool.resetForClinit(constantPoolIndex, constantPoolOffset);
        }
        else {
            if ((this.bits & 0x40) != 0x0) {
                int before = codeStream.position;
                codeStream.return_();
                if (sourcePosition != -1) {
                    codeStream.recordPositionsFrom(before, sourcePosition);
                }
            }
            codeStream.recordPositionsFrom(0, declaringType.sourceStart);
            classFile.completeCodeAttributeForClinit(codeAttributeOffset);
        }
    }
    
    @Override
    public boolean isClinit() {
        return true;
    }
    
    @Override
    public boolean isInitializationMethod() {
        return true;
    }
    
    @Override
    public boolean isStatic() {
        return true;
    }
    
    @Override
    public void parseStatements(Parser parser, CompilationUnitDeclaration unit) {
    }
    
    @Override
    public StringBuffer print(int tab, StringBuffer output) {
        ASTNode.printIndent(tab, output).append("<clinit>()");
        this.printBody(tab + 1, output);
        return output;
    }
    
    @Override
    public void resolve(ClassScope classScope) {
        this.scope = new MethodScope(classScope, classScope.referenceContext, true);
    }
    
    @Override
    public void traverse(ASTVisitor visitor, ClassScope classScope) {
        visitor.visit(this, classScope);
        visitor.endVisit(this, classScope);
    }
    
    public void setAssertionSupport(FieldBinding assertionSyntheticFieldBinding, boolean needClassLiteralField) {
        this.assertionSyntheticFieldBinding = assertionSyntheticFieldBinding;
        if (needClassLiteralField) {
            SourceTypeBinding sourceType = this.scope.outerMostClassScope().enclosingSourceType();
            if (!sourceType.isInterface() && !sourceType.isBaseType()) {
                this.classLiteralSyntheticField = sourceType.addSyntheticFieldForClassLiteral(sourceType, this.scope);
            }
        }
    }
}
